#ifndef __CLUTTER_DEPRECATED_H__
#define __CLUTTER_DEPRECATED_H__

#define __CLUTTER_DEPRECATED_H_INSIDE__

#include "deprecated/clutter-actor.h"
#include "deprecated/clutter-alpha.h"
#include "deprecated/clutter-animation.h"
#include "deprecated/clutter-box.h"
#include "deprecated/clutter-container.h"
#include "deprecated/clutter-group.h"
#include "deprecated/clutter-rectangle.h"
#include "deprecated/clutter-stage.h"
#include "deprecated/clutter-state.h"
#include "deprecated/clutter-timeline.h"

#undef __CLUTTER_DEPRECATED_H_INSIDE__

#endif /* __CLUTTER_DEPRECATED_H__ */
