/* -*- mode: C; c-file-style: "gnu"; indent-tabs-mode: nil; -*- */

#ifndef META_BACKGROUND_ACTOR_PRIVATE_H
#define META_BACKGROUND_ACTOR_PRIVATE_H

#include "meta/meta-background-actor.h"

cairo_region_t *meta_background_actor_get_clip_region (MetaBackgroundActor *self);

#endif /* META_BACKGROUND_ACTOR_PRIVATE_H */
